import numpy as np
import pprint as pp
import random as rand
from operator import attrgetter
from food_source import FoodSource
class ABC(object):
    """docstring for ABC"""
    food_sources = []
    gbest = []
    def __init__(self, npopulation, nruns, fn_eval, fn_lb, fn_ub, *, trials_limit=10, employed_bees_percentage=0.5):
        super(ABC, self).__init__()
        self.npopulation = npopulation
        self.nruns = nruns
        self.fn_eval = fn_eval
        self.trials_limit = trials_limit
        self.fn_lb = np.array(fn_lb)
        self.fn_ub = np.array(fn_ub)
        self.employed_bees = round(npopulation * employed_bees_percentage)
        self.onlooker_bees = npopulation - self.employed_bees
    def optimize(self):
        self.initialize()
        print("initial solution：")
        pp.pprint(self.food_sources)
        for nrun in range(1, self.nruns+1):
            self.employed_bees_stage()
            self.onlooker_bees_stage()
            self.scout_bees_stage()
        best_fs = self.best_source()
        print("Optimal timing parameters :")
        print(best_fs)
        return best_fs.solution
    def initialize(self):
        self.food_sources = [self.create_foodsource() for i in range(self.employed_bees)]
    def employed_bees_stage(self):
        # For each employed bee, a new solution is obtained by searching near the honey source, which is determined by greedy algorithm
        for i in range(self.employed_bees):
            food_source = self.food_sources[i]
            # Compare the new possible solution with the original solution and retain a better solution
            new_solution = self.generate_solution(i)
            new_solution = new_solution.astype(int)
            best_solution = self.best_solution(food_source.solution, new_solution)
            self.set_solution(food_source, best_solution)
    def onlooker_bees_stage(self):
        for i in range(self.onlooker_bees):
            #  Calculate the probability that the food source is selected by the following bee
            print(self.food_sources)
            probabilities = [self.probability(fs) for fs in self.food_sources]
            #  Follow the bee to hire the bee according to the roulette strategy (bee picking)
            #  According to the probability of each food source being selected, the food source is selected and the index number is stored in the selected_index
            print(probabilities)
            print(len(self.food_sources))
            selected_index = self.selection(range(len(self.food_sources)), probabilities)
            # Choose the food source
            selected_source = self.food_sources[selected_index]
            # To produce a new source of food near the source
            new_solution = self.generate_solution(selected_index)
            # Compare the new possible solution with the original solution and retain a better solution
            best_solution = self.best_solution(selected_source.solution, new_solution)
            print(best_solution)
            self.set_solution(selected_source, best_solution)
    def scout_bees_stage(self):
        # To determine if there's an abandoned honey source
        for i in range(self.employed_bees):
            food_source = self.food_sources[i]
            if food_source.trials > self.trials_limit:
                # Lead the peak into a scout bee and search for new food sources
                food_source = self.create_foodsource()
                self.food_sources[i] = food_source
    def generate_solution(self, current_solution_index):
        # solution represents the solution vector of the current index 
        solution = self.food_sources[current_solution_index].solution
        # randomly select a honey source whose index is not i 
        k_source_index = self.random_solution_excluding([current_solution_index])
        k_solution = self.food_sources[k_source_index].solution
        # Generate a new solution near the neighborhood of the honey source 
        d = rand.randint(0, len(self.fn_lb) - 1)
        r = rand.uniform(-1, 1)
        new_solution = np.copy(solution)
        r2 = rand.uniform(-1, 1)
        new_solution[d] = solution[d] + r * (solution[d] - k_solution[d])
        #new_solution[d] = solution[d] + r * (solution[d] - k_solution[d]) + r2 * (self.global_best[d]- solution[d] )
        # Generating integral solutions 
        new_solution[d] = new_solution[d].astype(int)
        while (new_solution[d] >= self.fn_ub[d]) or (new_solution[d] <= self.fn_lb[d]):
            #print("test1")
            print(new_solution[d])
            d = rand.randint(0, len(self.fn_lb) - 1)
            r = rand.uniform(-1, 1)
            new_solution = np.copy(solution)
            new_solution[d] = solution[d] + r * (solution[d] - k_solution[d])
            # Generating integral solutions 
            new_solution[d] = int(new_solution[d])
        return np.around(new_solution, decimals=4)
    def random_solution_excluding(self, excluded_index):
        available_indexes = set(range(self.employed_bees))
        exclude_set = set(excluded_index)
        diff = available_indexes - exclude_set
        selected = rand.choice(list(diff))
        return selected
    def best_solution(self, current_solution, new_solution):
        if self.fitness(new_solution) > self.fitness(current_solution):
            return new_solution
        else:
            return current_solution
    def probability(self, solution_fitness):
        fitness_sum = sum([fs.fitness for fs in self.food_sources])
        #print("fitness_sum:")
        print(fitness_sum)
        probability = solution_fitness.fitness / fitness_sum
        #print("cumulative probability")
        print(probability)
        return probability
    def fitness(self, solution):
        result = self.fn_eval(solution)
        #print("result:")
        #print(result)
        #fitness = 1 / (1 + result)
        if result > 0:
            fitness = 1 / (1 + result)
        else:
            fitness = 1 + abs(result)
        return fitness
    def selection(self, solutions, weights):
        print(rand.choices(solutions, weights))
        return rand.choices(solutions, weights)[0]
    def set_solution(self, food_source, new_solution):
        if np.array_equal(new_solution, food_source.solution):
            food_source.trials += 1
        else:
            food_source.solution = new_solution
            food_source.trials = 0
    def best_source(self):
        best = max(self.food_sources, key=attrgetter('fitness'))
        return best
    def create_foodsource(self):
        solution = self.candidate_solution(self.fn_lb, self.fn_ub)
        solution = solution.astype(int)
        fitness = self.fitness(solution)
        return FoodSource(solution, fitness)
    def candidate_solution(self, lb, ub):
        # Random generation of initial solutions 
        solution = [0 for i in range(len(ub))]
        for i in range(len(ub)):
            r = rand.random()
            solution[i] = lb[i] + (ub[i] - lb[i]) * r
        return np.around(solution, decimals=4)





