import math
import offset_funrange
from algorithm_swarm import ABC
from method_MPC import MPC
import win32com.client as com
vissim = com.Dispatch('VISSIM.Vissim.430')  #  Create vissim object
Filename = r"C:\Users\admin\Desktop\simulation20190214.inp"  #  Load map
net = vissim.Net
vissim.LoadNet(Filename)
sim = vissim.Simulation  # Create simulation objects
queuecounters = net.QueueCounters  # Defining queuing counter
vissim.Evaluation.SetAttValue("QUEUECOUNTER", True)  # Initialization property activation queue length evaluation
LinkNum = 14 #Number of road sections
cycle = 140 #Define period length is 140s
Desired_Speed = 50 * 1000 / 3600#car2(default vehicle speed limit), consistent with model predictive control methods
L = 200 # Section length all set is 200m
INTERSECTION = 2  #  Number of intersections
DIRECTION = 2  #  One way to 1, two way to 2
Var_size = (INTERSECTION-1) * DIRECTION  # Number of control variables
mpc = MPC()# Define a MPC object
print("hello")
class Delay(object):
    """docstring for delay calculate"""
    def __init__(self, time ,cycle =140, vfree = Desired_Speed, linkLength=L, mu = 1800):#Initialize properties
        self.time = time
        self.cycle = cycle
        self.vfree = vfree
        self.linklength =linkLength
        self.mu = mu # Section saturated traffic flow rate
    def offset_optimize(self):
        offset = self.optimize()# Obtained the phase difference optimization result
        return offset
    def optimize(self):
        INTERSECTION = 2  #  Number of intersections
        Var_size = INTERSECTION-1  # The number of control variables, determine where the relevant needs to be modified (phase difference only needs to determine the uplink or downlink phase difference, the other direction phase difference can be solved according to the known direction phase difference)
        offset_upper_bound, offset_lower_bound = offset_funrange.funRange(Var_size)#Obtain phase difference upper and lower bounds, set to 0~140s
        #abc_sphere2 = ABC(30, 49, self.delay_model,offset_lower_bound, offset_upper_bound)
        abc_sphere2 = ABC(40, 10, self.delay_model,offset_lower_bound, offset_upper_bound)
        offset = abc_sphere2.optimize()  # Gets the optimal solution of the delay model
        return offset
    def linkname_valid(self,i,j,k):#Determine whether the lane name is valid
        a = [2,4]
        b = [3,5]
        c = [1,7]
        d = [6,8]
        e = [3,4,5]
        f = [8,1,6]
        g = [6,7,8]
        h = [2,3,5]
        if (i in a) & (j == 1) & (k in b):
            flag = True
        elif (i in c) & (j == 2) & (k in d):
            flag = True
        elif (i in e) & (j == 1) & (k == 2):
            flag = True
        elif (i in f) & (j == 2) & (k == 7):
            flag = True
        elif (i in g) & (j == 2) & (k == 1):
            flag = True
        elif (i in h) & (j == 1) & (k == 4):
            flag = True
        else:
            flag = False
        print("Current section name L-%d-%d-%d"%(i,j,k))
        print(flag)
        return flag
    def delay_model(self,offset_down):# Expression of delay model
        queuelength = mpc.GetTotelQueueLength(sim.AttValue('ELAPSEDTIME'))  # Gets an array of queue lengths for all sections of the current sampling time
        #Delays in coordination direction
        delay_total = 0
        phase_no = 3
        # Delays in non-coordinated direction
        source_node = [1,2,3,4,5,6,7,8]#Two ends of the node
        inner_node = [1,2]#Internal nodes
        # Gets the current delay expression for all lanes
        for i in source_node:
            for j in inner_node:
                for k in source_node:
                    #If the lane name is valid, add the delay to the total delay calculation
                    if self.linkname_valid(i,j,k):
                        delay_total = delay_total + self.delay_calculate(i, j, k, offset_down,queuelength)
                    else:
                        delay_total = delay_total + 0
        return delay_total
    def delay_calculate(self,i,j,k,offset_down,queuelength):#Calculating delays in a lane
        #Determine the intersection phase of the corresponding traffic light according to the lane
        print(" Calculated lane name%d-%d-%d delay" % (i, j, k))
        [intersection_no,phase_no] = self.phase_no(i,j,k)
        print("phase_no为：%d"%phase_no)
        if (phase_no==3)& ((i != k-1) & (i != k-5))&(~((i==4)&(j==1)&(k==2)))&(~((i==7)&(j==2)&(k==1))) :#协调相位（第三相位非右转方向且不为协调相位第一个交叉口）/非协调相位判断
            delay = self.coordinate_delay(i, intersection_no, k, phase_no, offset_down, queuelength)
        else:
            delay = self.uncoordinate_delay(i, j, k, phase_no, queuelength)
        return delay
    def phase_no(self,i,j,k):#Determine the intersection and phase of the corresponding traffic light according to the lane
        intersection_no = j # Number of intersections
        #  To go straight or turn left in phase (first phase)
        a = [4, 2, 1, 7]  #  East - west intersection
        b = [3, 5, 6, 8]  # North-South junction
        if (i in a) & (k in b) & (i != k-1) & (i != k-5):  # First phase
            phase_no = 1
        elif ((i in a) & (k in a))| ((i in a) & (k in b) & (i != k+1) & (i != k-3)& (i != k-7)):  # Phase 3
            phase_no = 3
        else:  # 第二相位
            phase_no = 2
        return intersection_no,phase_no
    def coordinate_delay(self,i,intersection_no,k,phase_no,offset_down,queuelength):#Delays in coordination direction
        name, VNum = mpc.GetLinkImformation()  #  Get the number of vehicles in the corresponding section (the section evaluation is not given according to the per lane (set in the simulation software), otherwise the evaluation result is not the sampling result of a section)
        delay = 0
        # Determine the phase difference according to the direction of the upper and lower directions
        #Gets the phase difference currently to be calculated (Intersection2 the second element in the corresponding phase difference array)
        # Free travel time (from upstream intersection to downstream intersection)
        T = self.linklength / self.vfree
        # Downward directional phase difference acquisition
        if i < intersection_no:
            O_f = offset_down[intersection_no - 2]
            g_e = self.GetGreenTime(intersection_no - 1, phase_no)
        else:  # Upward directional phase difference acquisition
            O_f = self.cycle - offset_down[intersection_no - 2]
            g_e = self.GetGreenTime(intersection_no + 1, phase_no)
        #  Gets the green light timing of the coordination phase between the current intersection and the adjacent intersection
        f = intersection_no  #  Downstream intersection
        g_f = self.GetGreenTime(f, phase_no)
        #  Calculation of queuing time
        # Gets the remaining queue length of the previous cycle in the current direction (call detector detection data acquisition)
        H = self.GetResidualQueue(i, intersection_no, k, queuelength)
        # Red light time turns traffic to traffic
        V_R = self.GetUpRed_TimeFlow(queuelength, i, intersection_no, k, name, VNum)
        #V_R =1000
        #  Green time turns traffic to traffic
        V_G = self.GetUpGreen_TimeFlow(queuelength, i, intersection_no, k, name, VNum)
        #V_G =1000
        CT = (H + V_R * (self.cycle - g_e)) / self.mu
        # Total number of arrival vehicles calculated
        TA = V_G * g_e + V_R * (self.cycle - g_e)
        # Expressions for determining delays based on the relationship between G_e_i and adjacent intersections
        if CT + g_e <= g_f:
            delay = ((self.mu * (self.cycle - g_e - T + O_f) * (2 * H + V_R * (self.cycle - g_e - T + O_f)) + math.pow(H, 2)) / (self.mu - V_R)) / (2 * TA)
        elif CT < g_f <= CT + g_e:  #  In the second case, the range of phase difference needs to be further determined to determine the expression of delay
            if 0 <= O_f < T - CT:
                delay = ((self.mu * (self.cycle - g_e - T + O_f) * (2 * H + V_R * (self.cycle - g_e - T + O_f)) + math.pow(H, 2)) / (self.mu - V_R) + math.pow(g_e - g_f + T - O_f, 2) * V_G) / (2 * TA)
            elif T - CT <= O_f <= T:
                delay = ((self.cycle - g_e) * (V_R * (self.cycle - g_e) + 2 * H) - self.mu * math.pow(T - O_f,2) + V_G * math.pow(
                g_e - g_f + T - O_f, 2) + (self.mu * math.pow(V_R * (self.cycle - g_e) + H - self.mu * (T - O_f), 2)) / (self.mu - V_G)) / (2 * TA)
            elif T <= O_f <= T + g_e - g_f:
                delay = ((math.pow(V_R * (self.cycle - g_e) + H, 2) + self.mu * (O_f - T) * ((O_f - T) * V_G + 2 * V_R * (self.cycle - g_e) + 2 * H * (self.mu - g_e) * (self.mu - V_G))) / (self.mu - V_G) + V_G * math.pow(g_e - g_f + T - O_f, 2)) / (2 * TA)
        else:
            delay = ((self.cycle - g_e) * (V_R * (self.cycle + g_e) + 2 * H) + g_e * (g_e * V_R + 2 * H + 2 * V_R * (self.cycle + g_e)) - self.mu * g_f * (2 * g_e + 2 * T - 2 * O_f - g_f)) / (2 * TA)
        print("Delays in coordination direction:")
        print(delay)
        return delay
    def GetResidualQueue(self,i,j,k,queuelength):# Obtain the remaining queue length of the previous cycle in the current direction (consider simplifying and eliminating code duplication after optimization of cooperative phase difference and green light timing)
        queue = mpc.GetQueueLength(queuelength,i,j,k)
        return queue
    def GetUpRed_TimeFlow(self,queuelength,i,j,k,name, VNum):# Gets the arrival rate of the red light time at the upstream intersection
        if (i==1)&(j==2)&(k==7):#Downward coordination direction (second phase above auxiliary road left turn vehicle flow departure rate second phase below auxiliary road right turn vehicle flow departure rate)(3,1,2)and（5,1,2）
            alpha_arrival = mpc.GetAlphaLeave(queuelength, name, VNum, 3, 1, 2,self.time,0)+ mpc.GetAlphaLeave(queuelength, name, VNum, 5, 1, 2,self.time,0)
        if (i ==2) & (j ==1) & (k ==4):#Upward coordination direction（……）(6,2,1)and （8,2,1）
            #alpha_arrival = mpc.GetAlphaLeave(queuelength, name, VNum, 6, 2, 1,self.time,0)+ mpc.GetAlphaLeave(queuelength, name, VNum, 8, 2, 1,self.time,0)
            alpha_arrival = 1678
        print("Red light time attainment rate:")
        print(alpha_arrival)
        return alpha_arrival
    def GetUpGreen_TimeFlow(self,queuelength,i,j,k,name, VNum):  # Arrival rate of green time at upstream intersection
        if (i==1)&(j==2)&(k==7):#Downward coordination direction direct traffic departure rate(4,1,2)
            alpha_arrival = mpc.GetAlphaLeave(queuelength, name, VNum, 4, 1, 2,self.time,0)
        if (i ==2) & (j ==1) & (k ==4):# Uplink direction straight traffic departure rate(7,2,1)
            #alpha_arrival = mpc.GetAlphaLeave(queuelength, name, VNum, 7, 2, 1,self.time,0)
            alpha_arrival = mpc.GetAlphaArrive(queuelength, name, VNum, i,j,k,self.time, 0)
        print("Green Time Rate:")
        print(alpha_arrival)
        return alpha_arrival
    def uncoordinate_delay(self,i,intersection_no,k,phase_no,queuelength):# Delays in non-coordinated direction
        name, VNum = mpc.GetLinkImformation()  #  Calculation of non-coordinated direction delay to obtain the number of vehicles in the corresponding section (the evaluation of the section is not given according to the per lane (set in the simulation software), otherwise the evaluation result is not the sampling result of a section)
        #Gets the remaining queue length of the previous cycle in the current direction (call detector detection data acquisition)
        H = self.GetResidualQueue(i,intersection_no,k,queuelength)
        #H=1000
        #Getting green light at current intersection
        g_e = self.GetGreenTime(intersection_no,phase_no)
        V = self.GetFlowRate(i,intersection_no,k,queuelength,name,VNum,self.time)
        if g_e * self.mu > H + self.cycle * V:
            delay = (self.mu *(self.cycle -g_e)*(2*H + V*(self.cycle - g_e)) + math.pow(H,2))/ (2* self.cycle *(self.mu - V))
        else:
            delay = (self.cycle * (2 * H + V * self.cycle) - self.mu * math.pow(g_e,2)) / (2 * self.cycle * V)
        print("Delays in non-coordinating directions:")
        print(delay)
        return delay
    def GetFlowRate(self,i,j,k,queuelength,name, VNum,g):#Gets the departure rate in each direction (corresponding to the arrival rate of each phase in the original delay model)
        V = mpc.GetAlphaLeave(queuelength, name, VNum, i, j, k, g, 0)
        #V=1000
        return V
    def GetGreenTime(self,e,i):
        green_time = self.time[3*(e-1)+i-1]# Green time to get the corresponding phase of an intersection
        print(" Output green light at current intersection:")
        print(green_time)
        return green_time



